"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var Message_1 = require("../../Message");

var Symbols_1 = require("../../Symbols"); // typeof a !a()


function unaryExpressionHandler(node) {
  var _this = this;

  switch (node.operator) {
    case "delete":
      var objectGetter = this.createObjectGetter(node.argument);
      var nameGetter = this.createNameGetter(node.argument);
      return function () {
        // not allowed to delete root scope property
        // rootContext has move to prototype chai, so no judgment required
        // if (this.isRootScope(node.argument)) {
        // 	return false;
        // }
        var obj = objectGetter();
        var name = nameGetter();
        return delete obj[name];
      };

    default:
      var expression; // for typeof undefined var
      // typeof adf9ad

      if (node.operator === "typeof" && node.argument.type === "Identifier") {
        var _objectGetter = this.createObjectGetter(node.argument);

        var _nameGetter = this.createNameGetter(node.argument);

        expression = function expression() {
          return _objectGetter()[_nameGetter()];
        };
      } else {
        expression = this.createClosure(node.argument);
      }

      return function () {
        var value = expression();

        switch (node.operator) {
          case "-":
            return -value;

          case "+":
            return +value;

          case "!":
            return !value;

          case "~":
            return ~value;

          case "void":
            return void value;

          case "typeof":
            if (Symbols_1.isSymbol(value)) {
              return 'symbol';
            }

            return typeof value;

          default:
            throw _this.createInternalThrowError(Message_1.Messages.UnaryOperatorSyntaxError, node.operator, node);
        }
      };
  }
}

exports.unaryExpressionHandler = unaryExpressionHandler;