"use strict";

var _interopRequireDefault = require("@babel/runtime/helpers/interopRequireDefault");

var _inheritsLoose2 = _interopRequireDefault(require("@babel/runtime/helpers/inheritsLoose"));

Object.defineProperty(exports, "__esModule", {
  value: true
});

var Scope_1 = require("../Model/Scope");

var acorn_1 = require("acorn");

var Message_1 = require("../Model/Message");

var Closure_1 = require("../Model/Closure");

var Symbols_1 = require("../Model/Symbols");

var TokenClass_1 = require("../Model/TokenClass");

var util_1 = require("../util");

var Array_1 = require("../Model/Array");

var MyParser = acorn_1.Parser.extend(require('acorn-class-fields'), require('acorn-static-class-features'), require('acorn-jsx')());
var version = "1.1.1";

function internalFunction(reflection) {
  if (!(reflection instanceof InternalInterpreterReflection)) {
    throw new Error("Illegal call");
  }

  var instance = reflection.generator();

  for (var _len = arguments.length, params = new Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
    params[_key - 1] = arguments[_key];
  }

  var code = params.pop();
  var interpreter = new Interpreter(instance.getGlobalScope(), instance.getOptions());
  var wrapCode = "\n\t\t    (function anonymous(" + params.join(",") + "){\n\t\t        " + code + "\n\t\t    });\n\t\t    ";
  return interpreter.evaluate(wrapCode);
}

exports.internalFunction = internalFunction;
Object.defineProperty(internalFunction, "__IS_FUNCTION_FUNC", {
  value: true,
  writable: false,
  enumerable: false,
  configurable: false
});

var InternalInterpreterReflection =
/*#__PURE__*/
function () {
  function InternalInterpreterReflection(interpreter) {
    this.interpreter = interpreter;
  }

  var _proto = InternalInterpreterReflection.prototype;

  _proto.generator = function generator() {
    var interpreter = this.interpreter;

    function getCurrentScope() {
      return this.getCurrentScope();
    }

    function getGlobalScope() {
      return this.getGlobalScope();
    }

    function getCurrentContext() {
      return this.getCurrentContext();
    }

    return {
      getOptions: interpreter.getOptions.bind(interpreter),
      getCurrentScope: getCurrentScope.bind(interpreter),
      getGlobalScope: getGlobalScope.bind(interpreter),
      getCurrentContext: getCurrentContext.bind(interpreter),
      getExecStartTime: interpreter.getExecStartTime.bind(interpreter)
    };
  };

  return InternalInterpreterReflection;
}();

exports.InternalInterpreterReflection = InternalInterpreterReflection;

function internalEval(reflection, code, useGlobalScope) {
  if (useGlobalScope === void 0) {
    useGlobalScope = true;
  }

  if (!(reflection instanceof InternalInterpreterReflection)) {
    throw new Error("Illegal call");
  }

  if (typeof code !== "string") return code;
  if (!code) return void 0;
  var instance = reflection.generator();
  var opts = instance.getOptions();
  var options = {
    timeout: opts.timeout,
    _initEnv: function _initEnv() {
      // set caller context
      if (!useGlobalScope) {
        this.setCurrentContext(instance.getCurrentContext());
      } // share timeout


      this.execStartTime = instance.getExecStartTime();
      this.execEndTime = this.execStartTime;
    }
  };
  var currentScope = useGlobalScope ? instance.getGlobalScope() : instance.getCurrentScope();
  var interpreter = new Interpreter(currentScope, options);
  return interpreter.evaluate(code);
}

Object.defineProperty(internalEval, "__IS_EVAL_FUNC", {
  value: true,
  writable: false,
  enumerable: false,
  configurable: false
});
/**
 * scope chain
 *
 * superScope
 *     ↓
 * rootScope
 *     ↓
 * globalScope
 *     ↓
 * functionScope
 *
 */

function createRootContext(data) {
  return Object.create(data);
}

var BuildInObjects = {
  NaN: NaN,
  Infinity: Infinity,
  undefined: undefined,
  // null,
  Object: Object,
  // Array类是自定义的一个包装类，为了提供新特性的支持
  // Array,
  String: String,
  Boolean: Boolean,
  Number: Number,
  Date: Date,
  RegExp: RegExp,
  Error: Error,
  URIError: URIError,
  TypeError: TypeError,
  RangeError: RangeError,
  SyntaxError: SyntaxError,
  ReferenceError: ReferenceError,
  Math: Math,
  parseInt: parseInt,
  parseFloat: parseFloat,
  isNaN: isNaN,
  isFinite: isFinite,
  decodeURI: decodeURI,
  decodeURIComponent: decodeURIComponent,
  encodeURI: encodeURI,
  encodeURIComponent: encodeURIComponent,
  escape: escape,
  unescape: unescape,
  eval: internalEval,
  Function: internalFunction
}; // ES5 Object

if (typeof JSON !== "undefined") {
  BuildInObjects.JSON = JSON;
} //ES6 Object


if (typeof Promise !== "undefined") {
  BuildInObjects.Promise = Promise;
}

if (typeof Set !== "undefined") {
  BuildInObjects.Set = Set;
}

if (typeof Map !== "undefined") {
  BuildInObjects.Map = Map;
} // if (typeof Symbol !== "undefined") {
// 	BuildInObjects.Symbol = Symbol;
// }


if (typeof Proxy !== "undefined") {
  BuildInObjects.Proxy = Proxy;
}

if (typeof WeakMap !== "undefined") {
  BuildInObjects.WeakMap = WeakMap;
}

if (typeof WeakSet !== "undefined") {
  BuildInObjects.WeakSet = WeakSet;
}

if (typeof Reflect !== "undefined") {
  BuildInObjects.Reflect = Reflect;
}

var Interpreter =
/*#__PURE__*/
function (_Closure_1$ClosureHan) {
  (0, _inheritsLoose2.default)(Interpreter, _Closure_1$ClosureHan);

  function Interpreter(context, options) {
    var _this;

    if (context === void 0) {
      context = Interpreter.global;
    }

    if (options === void 0) {
      options = {};
    }

    _this = _Closure_1$ClosureHan.call(this) || this;
    _this.sourceList = []; // 编译时存放变量提升的var和function

    _this.collectDeclVars = Object.create(null);
    _this.collectDeclFuncs = Object.create(null); // 编译时存放块级作用域的层级结构的声明，用于判断var声明时候是否非法（var的申明可以覆盖var，但是不能覆盖let等块级声明）

    _this.collectDeclLex = [];
    _this.isVarDeclMode = false;
    _this.lastExecNode = null;
    _this.isRunning = false;
    _this.options = {
      ecmaVersion: options.ecmaVersion || Interpreter.ecmaVersion,
      timeout: options.timeout || 0,
      rootContext: options.rootContext,
      globalContextInFunction: options.globalContextInFunction === undefined ? Interpreter.globalContextInFunction : options.globalContextInFunction,
      _initEnv: options._initEnv
    };
    _this.context = context || Object.create(null);
    _this.callStack = []; // console.info('this i s ', this )

    _this.initEnvironment(_this.context);

    return _this;
  }
  /**
   * 初始化全局执行环境
   * @param ctx
   */


  var _proto2 = Interpreter.prototype;

  _proto2.initEnvironment = function initEnvironment(ctx) {
    var scope; //init global scope

    if (ctx instanceof Scope_1.default) {
      scope = ctx;
    } else {
      var rootScope = null;
      var superScope = this.createSuperScope(ctx);

      if (this.options.rootContext) {
        rootScope = new Scope_1.default(createRootContext(this.options.rootContext), superScope, Symbols_1.RootScopeName);
      }

      scope = new Scope_1.default(ctx, rootScope || superScope, Symbols_1.GlobalScopeName);
    }

    this.globalScope = scope; // 顶级作用域嵌入mock的Symbol方法。

    var smbl;
    this.globalScope.data['Symbol'] = smbl = Symbols_1.createSymbolFunc();
    this.globalScope.data['Array'] = Array_1.createArrayClass(smbl);
    this.currentScope = this.globalScope; //init global context to this

    this.globalContext = scope.data;
    this.currentContext = scope.data; // collect var/function declare

    this.collectDeclVars = Object.create(null);
    this.collectDeclFuncs = Object.create(null);
    this.collectDeclLex = [];
    this.execStartTime = Date.now();
    this.execEndTime = this.execStartTime;
    var _initEnv = this.options._initEnv;

    if (_initEnv) {
      _initEnv.call(this);
    }
  }
  /**
   * 包含块级作用域的语句声明之前执行。创建一个新的hash，记录块作用域内声明的let/const变量
   */
  ;

  _proto2.blockDeclareStart = function blockDeclareStart() {
    this.collectDeclLex.push(Object.create(null));
  }
  /**
   * 包含块级作用域的语句声明之后执行，退栈，获取该块作用域声明的let/const变量
   */
  ;

  _proto2.blockDeclareEnd = function blockDeclareEnd() {
    // 准备块级作用域初始化必要参数
    var lexDeclInThisBlock = this.collectDeclLex.pop();
    var lexDeclared;
    var lexNames = Object.getOwnPropertyNames(lexDeclInThisBlock);

    if (lexNames.length) {
      lexDeclared = Object.create(null);
      lexNames.forEach(function (key) {
        lexDeclared[key] = {
          init: false,
          kind: lexDeclInThisBlock[key].kind
        };
      });
    } else {
      // 说明没有词法变量，那就不用新建作用域了
      lexDeclared = null;
    }

    return lexDeclared;
  };

  _proto2.getExecStartTime = function getExecStartTime() {
    return this.execStartTime;
  };

  _proto2.getExecutionTime = function getExecutionTime() {
    return this.execEndTime - this.execStartTime;
  };

  _proto2.setExecTimeout = function setExecTimeout(timeout) {
    if (timeout === void 0) {
      timeout = 0;
    }

    this.options.timeout = timeout;
  };

  _proto2.getOptions = function getOptions() {
    return this.options;
  };

  _proto2.getGlobalScope = function getGlobalScope() {
    return this.globalScope;
  };

  _proto2.getCurrentScope = function getCurrentScope() {
    return this.currentScope;
  };

  _proto2.getCurrentContext = function getCurrentContext() {
    return this.currentContext;
  };

  _proto2.isInterruptThrow = function isInterruptThrow(err) {
    return err instanceof Message_1.InterruptThrowError || err instanceof Message_1.InterruptThrowReferenceError || err instanceof Message_1.InterruptThrowSyntaxError;
  };

  _proto2.createSuperScope = function createSuperScope(ctx) {
    var data = Object.assign({}, BuildInObjects);
    var buildInObjectKeys = Object.keys(data);
    buildInObjectKeys.forEach(function (key) {
      if (key in ctx) {
        delete data[key];
      }
    });
    return new Scope_1.default(data, null, Symbols_1.SuperScopeName);
  };

  _proto2.setCurrentContext = function setCurrentContext(ctx) {
    this.currentContext = ctx;
  };

  _proto2.setCurrentScope = function setCurrentScope(scope) {
    this.currentScope = scope;
  };

  _proto2.evaluate = function evaluate(code) {
    if (code === void 0) {
      code = "";
    }

    var node;
    if (!code) return;
    node = MyParser.parse(code, {
      ranges: true,
      locations: true,
      ecmaVersion: this.options.ecmaVersion || Interpreter.ecmaVersion
    });
    return this.evaluateNode(node, code);
  };

  _proto2.appendCode = function appendCode(code) {
    return this.evaluate(code);
  };

  _proto2.evaluateNode = function evaluateNode(node, source) {
    var _this2 = this;

    if (source === void 0) {
      source = "";
    }

    this.value = void 0;
    this.source = source;
    this.sourceList.push(source);
    this.isRunning = true; //reset timeout

    this.execStartTime = Date.now();
    this.execEndTime = this.execStartTime; // reset

    this.collectDeclVars = Object.create(null);
    this.collectDeclFuncs = Object.create(null);
    this.collectDeclLex = [];
    var currentScope = this.getCurrentScope();
    var currentContext = this.getCurrentContext();
    var labelStack = currentScope.labelStack.concat([]);
    var callStack = this.callStack.concat([]);

    var reset = function reset() {
      _this2.setCurrentScope(currentScope); //reset scope


      _this2.setCurrentContext(currentContext); //reset context


      currentScope.labelStack = labelStack; //reset label stack

      _this2.callStack = callStack; //reset call stack
    }; // start run


    try {
      var bodyClosure = this.createClosure(node); // add declares to data

      this.addDeclarationsToScope(this.collectDeclVars, this.collectDeclFuncs, this.getCurrentScope());
      var currScope = this.getCurrentScope();
      this._functionVarScope = currScope;
      bodyClosure();
    } catch (e) {
      throw e;
    } finally {
      reset();
      this.execEndTime = Date.now();
    }

    this.isRunning = false;
    return this.getValue();
  };

  _proto2.createErrorMessage = function createErrorMessage(msg, value, node) {
    var message = msg[1].replace("%0", String(value));

    if (node !== null) {
      message += this.getNodePosition(node || this.lastExecNode);
    }

    return message;
  };

  _proto2.createError = function createError(message, error) {
    return new error(message);
  };

  _proto2.createThrowError = function createThrowError(message, error) {
    return this.createError(message, error);
  };

  _proto2.createInternalThrowError = function createInternalThrowError(msg, value, node) {
    return this.createError(this.createErrorMessage(msg, value, node), msg[2]);
  };

  _proto2.checkTimeout = function checkTimeout() {
    if (!this.isRunning) return false;
    var timeout = this.options.timeout || 0;
    var now = Date.now();

    if (now - this.execStartTime > timeout) {
      return true;
    }

    return false;
  };

  _proto2.getNodePosition = function getNodePosition(node) {
    if (node) {
      var errorCode = ""; //this.source.slice(node.start, node.end);

      return node.loc ? " [" + node.loc.start.line + ":" + node.loc.start.column + "]" + errorCode : "";
    }

    return "";
  };

  _proto2.createClosure = function createClosure(node) {
    var _this3 = this;

    var closure = this.getClosure(node);

    if (!closure) {
      console.info('====> ', node);
      throw this.createInternalThrowError(Message_1.Messages.NodeTypeSyntaxError, node.type, node);
    }

    return function () {
      var timeout = _this3.options.timeout;

      if (timeout && timeout > 0 && _this3.checkTimeout()) {
        throw _this3.createInternalThrowError(Message_1.Messages.ExecutionTimeOutError, timeout, null);
      }

      _this3.lastExecNode = node;
      return closure.apply(void 0, arguments);
    };
  } // protected isRootScope(node: ESTree.Expression | ESTree.Pattern): boolean {
  // 	if (node.type === "Identifier") {
  // 		const scope = this.getScopeFromName(node.name, this.getCurrentScope());
  // 		return scope.name === "rootScope";
  // 	}
  // 	return false;
  // }
  ;

  _proto2.safeObjectGet = function safeObjectGet(obj, key, node) {
    if (Symbols_1.isSymbol(key)) {
      // 对于模拟的Symbol属性，实际存储在了一个特殊的key下面
      return obj[Symbols_1.storeKey(key)];
    }

    return obj[key];
  }
  /**
   * @param node
   */
  ;

  _proto2.createCallFunctionGetter = function createCallFunctionGetter(node) {
    var _this4 = this;

    switch (node.type) {
      case "MemberExpression":
        var objectGetter = this.createClosure(node.object);
        var keyGetter = this.createMemberKeyGetter(node);
        var source = this.source;
        return function () {
          var obj = objectGetter();
          var key = keyGetter();

          var func = _this4.safeObjectGet(obj, key, node);

          if (!func || !util_1.isFunction(func)) {
            var name = source.slice(node.start, node.end);
            throw _this4.createInternalThrowError(Message_1.Messages.FunctionUndefinedReferenceError, name, node);
          } // obj.eval = eval
          // obj.eval(...)


          if (func.__IS_EVAL_FUNC) {
            return function (code) {
              return func(new InternalInterpreterReflection(_this4), code, true);
            };
          } // obj.func = Function
          // obj.func(...)


          if (func.__IS_FUNCTION_FUNC) {
            return function () {
              for (var _len2 = arguments.length, args = new Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
                args[_key2] = arguments[_key2];
              }

              return func.apply(void 0, [new InternalInterpreterReflection(_this4)].concat(args));
            };
          } // method call
          // eg：obj.say(...)
          // eg: obj.say.call(...)
          // eg: obj.say.apply(...)
          // ======================
          // obj.func(...)
          // func = func.bind(obj)
          // tips:
          // func(...) -> func.bind(obj)(...)
          // func.call(...) -> obj.func.call.bind(obj.func)(...)
          // func.apply(...) -> obj.func.apply.bind(obj.func)(...)
          // ...others


          return func.bind(obj);
        };

      default:
        // test() or (0,test)() or a[1]() ...
        var closure = this.createClosure(node);
        return function () {
          var name = "";

          if (node.type === "Identifier") {
            name = node.name;
          } // console.info('the scope is ', this.getCurrentScope())
          // const name: string = (<ESTree.Identifier>node).name;


          var func = closure();

          if (!func || !util_1.isFunction(func)) {
            throw _this4.createInternalThrowError(Message_1.Messages.FunctionUndefinedReferenceError, name, node);
          } // https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/eval
          // var eval = eval;
          // function test(){
          //    eval(...); //note: use local scope in eval5，but in Browser is use global scope
          // }


          if (node.type === "Identifier" && func.__IS_EVAL_FUNC && name === "eval") {
            return function (code) {
              var scope = _this4.getScopeFromName(name, _this4.getCurrentScope());

              var useGlobalScope = scope.name === Symbols_1.SuperScopeName || // !scope.parent || // super scope
              scope.name === Symbols_1.GlobalScopeName || // this.globalScope === scope ||
              scope.name === Symbols_1.RootScopeName; // use local scope if calling eval in super scope

              return func(new InternalInterpreterReflection(_this4), code, !useGlobalScope);
            };
          } // use global scope
          // var g_eval = eval;
          // g_eval("a+1");
          //(0,eval)(...) ...eval alias


          if (func.__IS_EVAL_FUNC) {
            return function (code) {
              return func(new InternalInterpreterReflection(_this4), code, true);
            };
          } // Function('a', 'b', 'return a+b')


          if (func.__IS_FUNCTION_FUNC) {
            return function () {
              for (var _len3 = arguments.length, args = new Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
                args[_key3] = arguments[_key3];
              }

              return func.apply(void 0, [new InternalInterpreterReflection(_this4)].concat(args));
            };
          }

          var ctx = _this4.options.globalContextInFunction; // with(obj) {
          //     test() // test.call(obj, ...)
          // }

          if (node.type === "Identifier") {
            var scope = _this4.getIdentifierScope(node);

            if (scope.name === Symbols_1.WithScopeName) {
              ctx = scope.data;
            }
          } // function call
          // this = undefined
          // tips:
          // test(...) === test.call(undefined, ...)
          // fix: alert.call({}, ...) Illegal invocation


          return func.bind(ctx);
        };
    }
  };

  _proto2.getIdentifierScope = function getIdentifierScope(node) {
    var currentScope = this.getCurrentScope();
    var scope = this.getScopeFromName(node.name, currentScope);
    return scope;
  };

  _proto2.getVariableName = function getVariableName(node) {
    if (node.type === "Identifier") {
      return node.name;
    } else {
      throw this.createInternalThrowError(Message_1.Messages.VariableTypeSyntaxError, node.type, node);
    }
  }
  /**
   * 保证变量的访问只能在沙箱内部的。如果一个变量没有找到，会默认返回globalScope。
   * 那么如果globalScope如果不存在name的变量，就throw Error
   * @param data 作用域data
   * @param name Identifier标识符
   * @param node AST: Identifier Node
   */
  ;

  _proto2.assertVariable = function assertVariable(data, name, node) {
    if (data === this.globalScope.data && !(name in data)) {
      throw this.createInternalThrowError(Message_1.Messages.VariableUndefinedReferenceError, name, node);
    }
  } // get es3/5 param name
  ;

  _proto2.createParamNameGetter = function createParamNameGetter(node) {
    if (node.type === "Identifier") {
      return function () {
        return node.name;
      };
    } else if (node.type === 'RestElement') {
      return this.createParamNameGetter(node.argument);
    } else {
      throw this.createInternalThrowError(Message_1.Messages.ParamTypeSyntaxError, node.type, node);
    }
  };

  _proto2.createObjectKeyGetter = function createObjectKeyGetter(node) {
    var getter; // var obj = { title: "" }

    if (node.type === "Identifier") {
      getter = function getter() {
        return node.name;
      };
    } else {
      // Literal or ...
      // var obj = { "title": "" } or others...
      getter = this.createClosure(node);
    }

    return function () {
      return getter();
    };
  };

  _proto2.createMemberKeyGetter = function createMemberKeyGetter(node) {
    // s['a'];  node.computed = true
    // s.foo;  node.computed = false
    return node.computed ? this.createClosure(node.property) : this.createObjectKeyGetter(node.property);
  }
  /**
   * 如果一个变量作为左值调用，用这个Getter，主要是判断了const
   * 不是作为左值调用，用createObjectGetter即可
   * @param node
   */
  ;

  _proto2.createLeftObjectGetter = function createLeftObjectGetter(node) {
    var _this5 = this;

    switch (node.type) {
      case 'Identifier':
        return function () {
          var name = node.name;

          var scope = _this5.getScopeFromName(name, _this5.getCurrentScope(), true);

          if (scope.lexDeclared && scope.lexDeclared[name] && scope.lexDeclared[name].kind == 'const') {
            if (scope.lexDeclared[name].init === false) {
              scope.lexDeclared[name].init = true;
            } else {
              // console.info('node ====> ', node, scope)
              throw _this5.createInternalThrowError(Message_1.Messages.ConstChangeError, name, node);
            }
          }

          return scope.data;
        };

      case "MemberExpression":
        return this.createClosure(node.object);

      default:
        throw this.createInternalThrowError(Message_1.Messages.AssignmentTypeSyntaxError, node.type, node);
    }
  } // for UnaryExpression UpdateExpression AssignmentExpression
  ;

  _proto2.createObjectGetter = function createObjectGetter(node) {
    var _this6 = this;

    switch (node.type) {
      case "Identifier":
        return function () {
          return _this6.getScopeDataFromName(node.name, _this6.getCurrentScope());
        };

      case "MemberExpression":
        return this.createClosure(node.object);

      default:
        throw this.createInternalThrowError(Message_1.Messages.AssignmentTypeSyntaxError, node.type, node);
    }
  } // for UnaryExpression UpdateExpression AssignmentExpression
  ;

  _proto2.createNameGetter = function createNameGetter(node) {
    switch (node.type) {
      case "Identifier":
        return function () {
          return node.name;
        };

      case "MemberExpression":
        return this.createMemberKeyGetter(node);

      default:
        throw this.createInternalThrowError(Message_1.Messages.AssignmentTypeSyntaxError, node.type, node);
    }
  } // 创建closure的时候就会注册变量，等于是完成了变量和函数声明的提前操作
  ;

  _proto2.varDeclaration = function varDeclaration(name) {
    var context = this.collectDeclVars;
    context[name] = undefined;
  };

  _proto2.funcDeclaration = function funcDeclaration(name, func) {
    var context = this.collectDeclFuncs;
    context[name] = func;
  }
  /**
   * 在当前词法作用域下声明一个词法变量。这里的声明在词法作用域下是提升的，为的是鉴别未初始化之前对词法变量的访问。
   * 如果一个词法变量init:false,说明这个词法变量在该作用域内是存在的，但是还未初始化，访问它就是非法的。
   * @param name
   */
  // protected lextDeclaration(name: string): void {
  // 	// 声明的时候只可能出现在一个块级作用域的初始状态
  // 	let blockContext = this.lexContext[this.lexContext.length - 1]
  // 	blockContext[name] = {
  // 		init: false,
  // 		val: void 0,
  // 	}
  // }

  /**
   * 将实现扫描得到的变量声明和函数申明提前放到scope.data中
   * 注意变量的声明放进去的值只是undefined
   * @param declVars
   * @param declFuncs
   * @param scope
   */
  ;

  _proto2.addDeclarationsToScope = function addDeclarationsToScope(declVars, declFuncs, scope) {
    var scopeData = scope.data;

    for (var key in declFuncs) {
      var value = declFuncs[key];
      scopeData[key] = value ? value() : value;
    }

    for (var _key4 in declVars) {
      if (!(_key4 in scopeData)) {
        scopeData[_key4] = void 0;
      }
    }
  };

  _proto2.getScopeValue = function getScopeValue(name, startScope) {
    var scope = this.getScopeFromName(name, startScope);
    return scope.data[name];
  };

  _proto2.getScopeDataFromName = function getScopeDataFromName(name, startScope) {
    return this.getScopeFromName(name, startScope).data;
  }
  /**
   * 从当前scope从下往上找，如果当前scope.type=='block'，
   * 先确保scope.lexDeclared[name]是否为false，
   * 如果为false，说明这个块变量没有初始化，直接报错就行了
   * 如果为true，再去scope.data里面找
   * 如果scope.type=='function'，直接去scope.data里面找。
   * @param name
   * @param startScope
   */
  ;

  _proto2.getScopeFromName = function getScopeFromName(name, startScope, constInit) {
    var scope = startScope;

    do {
      if (scope.type == 'block') {
        if (!scope.lexDeclared[name]) {
          // 按理说，一个blockscope上的所有变量都标记在lexDeclared中，也有例外情况，
          // 比如catch(e){}中，变量e会零时插入，且lexDeclared上也不会有标记
          if (name in scope.data) {
            return scope;
          } // 否则向上找

        } else {
          if (scope.lexDeclared[name].init === false) {
            if (constInit && scope.lexDeclared[name].kind == 'const') {
              // const变量初始化的时候这个校验位还没有置为true，放行
              return scope;
            }

            throw this.createInternalThrowError(Message_1.Messages.LetVariableUseBeforeInitReferenceError, name);
          } else if (scope.lexDeclared[name].init === true) {
            return scope;
          } // 否则向上找

        }
      } else {
        // function scope
        // 函数级别的作用域也可能有lexDeclared
        if (scope.lexDeclared && scope.lexDeclared[name] && scope.lexDeclared[name].init === false) {
          throw this.createInternalThrowError(Message_1.Messages.LetVariableUseBeforeInitReferenceError, name);
        }

        if (name in scope.data) {
          //if (hasOwnProperty.call(scope.data, name)) {
          return scope;
        } // 否则向上找

      }
    } while (scope = scope.parent);

    return this.globalScope;
  };

  _proto2.entryBlockScope = function entryBlockScope(newScope) {
    var prevScope = this.getCurrentScope(); // 函数执行时，创建新的scope，然后下一行将程序的运行指针指向新的scope

    this.setCurrentScope(newScope); // blockScope不用赋值新的提升变量
    // self.addDeclarationsToScope(declVars, declFuncs, currentScope);

    return prevScope;
  };

  _proto2.setValue = function setValue(value) {
    var isFunctionCall = this.callStack.length;

    if (this.isVarDeclMode || isFunctionCall || value === Symbols_1.EmptyStatementReturn || value === Symbols_1.Break || value === Symbols_1.Continue || value instanceof TokenClass_1.BreakLabel || value instanceof TokenClass_1.ContinueLabel) {
      return value;
    }

    this.value = value instanceof TokenClass_1.Return ? value.value : value;
    return value;
  };

  _proto2.getValue = function getValue() {
    return this.value;
  };

  return Interpreter;
}(Closure_1.ClosureHandler);

exports.Interpreter = Interpreter;
Interpreter.version = version;
Interpreter.eval = internalEval;
Interpreter.Function = internalFunction; // 默认切换到2018，支持解构

Interpreter.ecmaVersion = 2018; // alert.call(globalContextInFunction, 1);
// fix: alert.call({}, 1); // Illegal invocation
// function func(){
//     this;// Interpreter.globalContextInFunction
// }
// func()

Interpreter.globalContextInFunction = void 0;
Interpreter.global = Object.create(null);