"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});

var Message_1 = require("../../Message");

var react_1 = require("../../Node/react"); // <div>hello</div>


function JSXElementHandler(node) {
  var _this = this;

  var React = this.globalContext['React'];

  if (!React) {
    throw this.createInternalThrowError(Message_1.Messages.NormalError, 'to use jsx, you must set React in globalContext', node);
  }

  var tag = node.openingElement.name;
  var tagClosure;

  if (tag.type == 'JSXIdentifier') {
    var tagname = tag.name;
    tagClosure = react_1.REACT_INTRINSIC_TAG.includes(tag.name) ? // 直接返回内置标签名
    function () {
      return tagname;
    } // 找到这个变量
    : this.createClosure({
      type: 'Identifier',
      name: tagname,
      loc: tag.loc
    });
  } else if (tag.type == 'JSXMemberExpression') {
    // @ts-ignore
    tagClosure = this.createClosure(tag);
  } else if (tag.type == 'JSXNamespacedName') {
    throw this.createInternalThrowError(Message_1.Messages.NormalError, 'jsx namespacename not support. ', node);
  }

  var childrenClosure;

  if (node.openingElement.selfClosing || !node.children || !node.children.length) {
    childrenClosure = [function () {
      return null;
    }];
  } else {
    childrenClosure = node.children.map(function (item) {
      if (item.type == 'JSXElement') {
        return _this.createClosure(item);
      } else if (item.type == 'JSXExpressionContainer') {
        return _this.createClosure(item.expression);
      } else if (item.type == 'JSXText') {
        var val = item.value.trim();
        return function () {
          return val;
        };
      } else {
        // fragment和spreadChild不支持
        return function () {
          return null;
        };
      }
    });
  }

  var attrClosure;

  if (!node.openingElement.attributes || !node.openingElement.attributes.length) {
    attrClosure = [function () {
      return null;
    }];
  } else {
    attrClosure = node.openingElement.attributes.map(function (item) {
      return _this.createClosure(item);
    });
  }

  return function () {
    var tag = tagClosure();
    var attrs = attrClosure.map(function (_) {
      return _();
    }).filter(function (_) {
      return _;
    });
    var attr = attrs.length ? Object.assign.apply({}, attrs) : null;
    var childs = childrenClosure.map(function (_) {
      return _();
    }).filter(function (_) {
      return _;
    });
    return React.createElement.apply(React, [tag, attr].concat(childs));
  };
}

exports.JSXElementHandler = JSXElementHandler;