"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Message_1 = require("../../Message");
const Symbols_1 = require("../../Symbols");
// typeof a !a()
function unaryExpressionHandler(node) {
    switch (node.operator) {
        case "delete":
            const objectGetter = this.createObjectGetter(node.argument);
            const nameGetter = this.createNameGetter(node.argument);
            return () => {
                // not allowed to delete root scope property
                // rootContext has move to prototype chai, so no judgment required
                // if (this.isRootScope(node.argument)) {
                // 	return false;
                // }
                let obj = objectGetter();
                const name = nameGetter();
                return delete obj[name];
            };
        default:
            let expression;
            // for typeof undefined var
            // typeof adf9ad
            if (node.operator === "typeof" && node.argument.type === "Identifier") {
                const objectGetter = this.createObjectGetter(node.argument);
                const nameGetter = this.createNameGetter(node.argument);
                expression = () => objectGetter()[nameGetter()];
            }
            else {
                expression = this.createClosure(node.argument);
            }
            return () => {
                const value = expression();
                switch (node.operator) {
                    case "-":
                        return -value;
                    case "+":
                        return +value;
                    case "!":
                        return !value;
                    case "~":
                        return ~value;
                    case "void":
                        return void value;
                    case "typeof":
                        if (Symbols_1.isSymbol(value)) {
                            return 'symbol';
                        }
                        return typeof value;
                    default:
                        throw this.createInternalThrowError(Message_1.Messages.UnaryOperatorSyntaxError, node.operator, node);
                }
            };
    }
}
exports.unaryExpressionHandler = unaryExpressionHandler;
