"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Scope_1 = require("../../Scope");
const TokenClass_1 = require("../../TokenClass");
const util_1 = require("../../../util");
// var f = function() {...}
// 在es5的标准下，没有块级作用域，只会在function下会创建一个新的scope。所以这里的操作和evaluate的感觉差不多
function arrowFunctionExpressionHandler(node) {
    const source = this.source;
    const oldDeclVars = this.collectDeclVars;
    const oldDeclFuncs = this.collectDeclFuncs;
    const oldDeclLex = this.collectDeclLex;
    // 准备新scope的声明提升变量
    this.collectDeclVars = Object.create(null);
    this.collectDeclFuncs = Object.create(null);
    this.collectDeclLex = [];
    const name = 'anonymous_arrow_func'; /**anonymous*/
    const paramLength = node.params.length;
    const paramsGetter = node.params.map(param => ({
        type: param.type,
        closure: this.createParamNameGetter(param),
    }));
    this.blockDeclareStart();
    // set scope
    // @ts-ignore
    const bodyClosure = this.createClosure(node.expression === true ? {
        type: 'GroupStatement',
        body: [{
                type: 'ReturnStatement',
                argument: node.body,
            }]
    } : node.body);
    let lexDecls = this.blockDeclareEnd();
    // 这里是准备好的变量和函数声明提升
    const declVars = this.collectDeclVars;
    const declFuncs = this.collectDeclFuncs;
    const declLex = this.collectDeclLex;
    this.collectDeclVars = oldDeclVars;
    this.collectDeclFuncs = oldDeclFuncs;
    this.collectDeclLex = oldDeclLex;
    // 创建一个新的scope，然后返回一个函数，该函数会在新建的scope执行
    return () => {
        // bind current scope
        // 箭头函数的scope和context都是在函数声明的时候就确定了。
        const runtimeScope = this.getCurrentScope();
        const ctx = this.getCurrentContext();
        const func = (...args) => {
            this.callStack.push(`${name}`);
            const prevScope = this.getCurrentScope();
            // 函数执行时，创建新的scope，然后下一行将程序的运行指针指向新的scope
            const currentScope = Scope_1.createScope(runtimeScope, `FunctionScope(${name})`);
            currentScope.lexDeclared = lexDecls;
            this.setCurrentScope(currentScope);
            // 将准备好的变量和函数声明赋值到新的scope
            this.addDeclarationsToScope(declVars, declFuncs, currentScope);
            // var t = function(){ typeof t } // function
            // t = function(){ typeof t } // function
            // z = function tx(){ typeof tx } // function
            // but
            // d = { say: function(){ typeof say } } // undefined
            // arrow-function has no 'arguments'
            // currentScope.data["arguments"] = arguments;
            paramsGetter.forEach((getter, i) => {
                if (getter.type === 'RestElement') {
                    currentScope.data[getter.closure()] = args.slice(i);
                }
                else {
                    currentScope.data[getter.closure()] = args[i];
                }
            });
            // init this
            const prevContext = this.getCurrentContext();
            //for ThisExpression
            this.setCurrentContext(ctx);
            // 执行
            const result = bodyClosure();
            // 恢复
            this.setCurrentContext(prevContext);
            this.setCurrentScope(prevScope);
            this.callStack.pop();
            if (result instanceof TokenClass_1.Return) {
                return result.value;
            }
        };
        // todo: let foo = ()=>{}, foo.name==='foo'
        util_1.defineFunctionName(func, name);
        // arrow-func has no .length
        // Object.defineProperty(func, "length", {
        //     value: paramLength,
        //     writable: false,
        //     enumerable: false,
        //     configurable: true,
        // });
        Object.defineProperty(func, "toString", {
            value: () => {
                return source.slice(node.start, node.end);
            },
            writable: true,
            configurable: true,
            enumerable: false,
        });
        Object.defineProperty(func, "valueOf", {
            value: () => {
                return source.slice(node.start, node.end);
            },
            writable: true,
            configurable: true,
            enumerable: false,
        });
        return func;
    };
}
exports.arrowFunctionExpressionHandler = arrowFunctionExpressionHandler;
