"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const util_1 = require("../../../util");
const Symbols_1 = require("../../Symbols");
const TokenClass_1 = require("../../TokenClass");
const Scope_1 = require("../../Scope");
// for(var i = 0; i < 10; i++) {...}
function forStatementHandler(node) {
    let initClosure = util_1.noop;
    let testClosure = node.test ? this.createClosure(node.test) : () => true;
    let updateClosure = util_1.noop;
    let initLexDeclared;
    let getBodyClosure = () => {
        if (node.body.type == 'BlockStatement') {
            // node.body是个blockStatement，这里我们就不多此一举再去检测块作用域了。
            return {
                needBlock: null,
                closure: this.createClosure(node.body)
            };
        }
        else {
            this.blockDeclareStart();
            let closure = this.createClosure(node.body);
            let bodyLex = this.blockDeclareEnd();
            return {
                needBlock: bodyLex,
                closure,
            };
        }
    };
    if (node.type === "ForStatement") {
        if (node.init) {
            this.blockDeclareStart();
            initClosure = this.createClosure(node.init);
            initLexDeclared = this.blockDeclareEnd();
        }
        updateClosure = node.update ? this.createClosure(node.update) : util_1.noop;
    }
    else {
        // while, do-while
    }
    return pNode => {
        let labelName;
        let result = Symbols_1.EmptyStatementReturn;
        let shouldInitExec = node.type === "DoWhileStatement";
        if (pNode && pNode.type === "LabeledStatement") {
            labelName = pNode.label.name;
        }
        let prevScope;
        let newScope;
        if (initLexDeclared) {
            newScope = Scope_1.createScope(this.getCurrentScope(), `BScope(for-let)`, "block");
            newScope.lexDeclared = initLexDeclared;
            prevScope = this.entryBlockScope(newScope);
        }
        for (initClosure(); shouldInitExec || testClosure(); updateClosure()) {
            shouldInitExec = false;
            let bodyClosure = getBodyClosure();
            let bodyPrev;
            let bodyScope;
            if (bodyClosure.needBlock) {
                bodyScope = Scope_1.createScope(this.getCurrentScope(), `BScope(for-body)`, "block");
                bodyScope.lexDeclared = bodyClosure.needBlock;
                bodyPrev = this.entryBlockScope(bodyScope);
            }
            // save last value
            const ret = this.setValue(bodyClosure.closure());
            // 恢复作用域
            if (bodyClosure.needBlock) {
                this.setCurrentScope(bodyPrev);
            }
            // notice: never return Break or Continue!
            if (ret === Symbols_1.EmptyStatementReturn || ret === Symbols_1.Continue)
                continue;
            if (ret === Symbols_1.Break) {
                break;
            }
            result = ret;
            // stop continue label
            if (result instanceof TokenClass_1.ContinueLabel && result.value === labelName) {
                result = Symbols_1.EmptyStatementReturn;
                continue;
            }
            if (result instanceof TokenClass_1.Return ||
                result instanceof TokenClass_1.BreakLabel ||
                result instanceof TokenClass_1.ContinueLabel) {
                break;
            }
        }
        if (initLexDeclared) {
            this.setCurrentScope(prevScope);
        }
        return result;
    };
}
exports.forStatementHandler = forStatementHandler;
