"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Symbols_1 = require("../../Symbols");
const TokenClass_1 = require("../../TokenClass");
const Scope_1 = require("../../Scope");
function switchStatementHandler(node) {
    const discriminantClosure = this.createClosure(node.discriminant);
    this.blockDeclareStart();
    const caseClosures = node.cases.map(item => {
        const testClosure = item.test ? this.createClosure(item.test) : () => Symbols_1.DefaultCase;
        const bodyClosure = this.createClosure({
            // @ts-ignore
            type: "GroupStatement", body: item.consequent,
        });
        return () => ({
            testClosure,
            bodyClosure,
        });
    });
    let blockLexDecl = this.blockDeclareEnd();
    return () => {
        const value = discriminantClosure();
        let match = false;
        let result;
        let ret, defaultCase;
        // switch执行的时候共用同一个blockscope
        let prevScope;
        let newScope;
        if (blockLexDecl) {
            newScope = Scope_1.createScope(this.getCurrentScope(), `BScope(switch)`, "block");
            newScope.lexDeclared = blockLexDecl;
            prevScope = this.entryBlockScope(newScope);
        }
        for (let i = 0; i < caseClosures.length; i++) {
            const item = caseClosures[i]();
            const test = item.testClosure();
            if (test === Symbols_1.DefaultCase) {
                defaultCase = item;
                continue;
            }
            if (match || test === value) {
                match = true;
                ret = this.setValue(item.bodyClosure());
                // notice: never return Break!
                if (ret === Symbols_1.EmptyStatementReturn)
                    continue;
                if (ret === Symbols_1.Break) {
                    break;
                }
                result = ret;
                if (result instanceof TokenClass_1.Return ||
                    result instanceof TokenClass_1.BreakLabel ||
                    result instanceof TokenClass_1.ContinueLabel ||
                    result === Symbols_1.Continue) {
                    break;
                }
            }
        }
        if (!match && defaultCase) {
            ret = this.setValue(defaultCase.bodyClosure());
            const isEBC = ret === Symbols_1.EmptyStatementReturn || ret === Symbols_1.Break || ret === Symbols_1.Continue;
            // notice: never return Break or Continue!
            if (!isEBC) {
                result = ret;
            }
        }
        if (blockLexDecl) {
            this.setCurrentScope(prevScope);
        }
        return result;
    };
}
exports.switchStatementHandler = switchStatementHandler;
